/* save.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



/* Graphics Save & Restore:

   The file is a list of tagged blocks.  A block starts with
   a tag and length.  If the restore function finds an unknown tag, it
   can skip the block and continue.  This provides for backward
   compatibility with future formats.
*/


#include <stdio.h>
#include "globals.h"
#include "memory.h"
#include "sync.h"
#include "topo.h"


#define INT_1_SIZE  sizeof(int_1)
#define UINT_1_SIZE sizeof(uint_1)
#define INT_2_SIZE  sizeof(int_2)
#define UINT_2_SIZE sizeof(uint_2)
#define INT_SIZE    sizeof(int)
#define FLOAT_SIZE  sizeof(float)
#define ULONG_SIZE  sizeof(unsigned long)


#define TAG_SIZE 4

#define TAG_ISOSURF  1
#define TAG_HSLICE   2
#define TAG_VSLICE   3
#define TAG_CHSLICE  4
#define TAG_CVSLICE  5
#define TAG_TRAJ     6
#define TAG_HWIND    8
#define TAG_VWIND    9

#define TAG_ISO_COLOR      11
#define TAG_HSLICE_COLOR   12
#define TAG_VSLICE_COLOR   13
#define TAG_CHSLICE_COLORS 14
#define TAG_CVSLICE_COLORS 15
#define TAG_TRAJ_COLOR     16
#define TAG_HWIND_COLOR    18
#define TAG_VWIND_COLOR    19

#define TAG_CTM          30
#define TAG_LABEL        31
#define TAG_TOPO_COLORS  32
#define TAG_CLONE_TABLE  33


#define FWRITE( BUF, SIZE, ITEMS, F )  if (fwrite(BUF,SIZE,ITEMS,F)!=ITEMS) { \
                                          fclose(F);              \
					  LOCK_OFF( Gfx_Lock );   \
                                          return 2;               \
				       }



/* skip 'n' bytes in file 'f' */
static void skip( f, n )
FILE *f;
int n;
{
   int i;

   for (i=0;i<n;i++)
      fgetc( f );
}




/*** save *************************************************************
   Save current graphics and colors to the 'savefile'
   Input:  savefile - filename to save to.
           justcolors - if non-zero, just save colors
   Return:  0 for success,
            1 if unable to open file
	    2 if error while writing file.
**********************************************************************/
int save( savefile, justcolors )
char savefile[];
int justcolors;
{
   FILE *f;
   struct label *lab;
   int tag, it, ip;
   int i, j, bytes, size, num1, num2, num3;

   f = fopen(savefile,"w");
   if (!f)
      return 1;


   LOCK_ON( Gfx_Lock );

   /* save clone table */
   tag = TAG_CLONE_TABLE;
   FWRITE( &tag, TAG_SIZE, 1, f );
   i = NumParms + NumClones;
   bytes = INT_SIZE + INT_SIZE*i;
   FWRITE( &bytes, INT_SIZE, 1, f );
   FWRITE( &i, INT_SIZE, 1, f );
   FWRITE( CloneTable, INT_SIZE, i, f );

   if (!justcolors) {
      /* save isosurfaces */
      for (ip=0;ip<NumParms+NumClones;ip++) {
	 for (it=0;it<NumTimes;it++) {
	    if (SurfTable[ip][it].valid) {
	       int numverts, numpts;
	       tag = TAG_ISOSURF;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       numverts = SurfTable[ip][it].numverts;
	       numpts = SurfTable[ip][it].numpts;
	       bytes = 4*INT_SIZE + FLOAT_SIZE
		     + numverts*(3*INT_2_SIZE+3*INT_1_SIZE)
		     + numpts*UINT_2_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       /* isosurface data */
	       FWRITE( &ip, INT_SIZE, 1, f );  /* parm */
	       FWRITE( &it, INT_SIZE, 1, f );  /* time */
	       FWRITE( &SurfTable[ip][it].isolevel, FLOAT_SIZE, 1, f );
	       FWRITE( &numverts, INT_SIZE, 1, f );  /* number of vertices */
	       FWRITE( &numpts, INT_SIZE, 1, f );
	       FWRITE( SurfTable[ip][it].vx, INT_2_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].vy, INT_2_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].vz, INT_2_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].nx, INT_1_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].ny, INT_1_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].nz, INT_1_SIZE, numverts, f );
	       FWRITE( SurfTable[ip][it].pts, UINT_2_SIZE, numpts, f );
	    }
	 }
      }
   }

   /* save isosurface colors */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      tag = TAG_ISO_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, 4, 1, f );
      FWRITE( &ip, INT_SIZE, 1, f );
      FWRITE( &Color[ip][ISOSURF], ULONG_SIZE, 1, f );
   }

   if (!justcolors) {
      /* save horizontal contour line slices */
      for (ip=0;ip<NumParms+NumClones;ip++) {
	 for (it=0;it<NumTimes;it++) {
	    if (HSliceTable[ip][it].valid) {
	       tag = TAG_HSLICE;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       num1 = HSliceTable[ip][it].num1;
	       num2 = HSliceTable[ip][it].num2;
	       num3 = HSliceTable[ip][it].num3;
	       bytes = 7*INT_SIZE + (num1+num2+num3) * 2 * INT_2_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       /* contour line data */
	       FWRITE( &ip, INT_SIZE, 1, f );  /* parm */
	       FWRITE( &it, INT_SIZE, 1, f );  /* time */
	       FWRITE( &HSliceTable[ip][it].interval, FLOAT_SIZE, 1, f );
	       FWRITE( &HSliceTable[ip][it].level, FLOAT_SIZE, 1, f );
	       FWRITE( &num1, INT_SIZE, 1, f );
	       FWRITE( HSliceTable[ip][it].vx1, INT_2_SIZE, num1, f );
	       FWRITE( HSliceTable[ip][it].vy1, INT_2_SIZE, num1, f );
	       FWRITE( &num2, INT_SIZE, 1, f );
	       FWRITE( HSliceTable[ip][it].vx2, INT_2_SIZE, num2, f );
	       FWRITE( HSliceTable[ip][it].vy2, INT_2_SIZE, num2, f );
	       FWRITE( &num3, INT_SIZE, 1, f );
	       FWRITE( HSliceTable[ip][it].vx3, INT_2_SIZE, num3, f );
	       FWRITE( HSliceTable[ip][it].vy3, INT_2_SIZE, num3, f );
	    }
	 }
      }
   }
      
   /* Save horizontal contour line slice colors */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      tag = TAG_HSLICE_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      FWRITE( &ip, INT_SIZE, 1, f );
      FWRITE( &Color[ip][HSLICE], ULONG_SIZE, 1, f );
   }
   
   if (!justcolors) {
      /* save vertical contour slices */
      for (ip=0;ip<NumParms+NumClones;ip++) {
	 for (it=0;it<NumTimes;it++) {
	    if (VSliceTable[ip][it].valid) {
	       tag = TAG_VSLICE;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       num1 = VSliceTable[ip][it].num1;
	       num2 = VSliceTable[ip][it].num2;
	       num3 = VSliceTable[ip][it].num3;
	       bytes = 5*INT_SIZE + 5*FLOAT_SIZE
		     + (num1+num2+num3) * 3 * INT_2_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       /* contour line data */
	       FWRITE( &ip, INT_SIZE, 1, f );  /* parm */
	       FWRITE( &it, INT_SIZE, 1, f );  /* time */
	       FWRITE( &VSliceTable[ip][it].interval, FLOAT_SIZE, 1, f );
	       FWRITE( &VSliceTable[ip][it].r1, FLOAT_SIZE, 1, f );
	       FWRITE( &VSliceTable[ip][it].c1, FLOAT_SIZE, 1, f );
	       FWRITE( &VSliceTable[ip][it].r2, FLOAT_SIZE, 1, f );
	       FWRITE( &VSliceTable[ip][it].c2, FLOAT_SIZE, 1, f );
	       FWRITE( &num1, INT_SIZE, 1, f );
	       FWRITE( VSliceTable[ip][it].vx1, INT_2_SIZE, num1, f );
	       FWRITE( VSliceTable[ip][it].vy1, INT_2_SIZE, num1, f );
	       FWRITE( VSliceTable[ip][it].vz1, INT_2_SIZE, num1, f );
	       FWRITE( &num2, INT_SIZE, 1, f );
	       FWRITE( VSliceTable[ip][it].vx2, INT_2_SIZE, num2, f );
	       FWRITE( VSliceTable[ip][it].vy2, INT_2_SIZE, num2, f );
	       FWRITE( VSliceTable[ip][it].vz2, INT_2_SIZE, num2, f );
	       FWRITE( &num3, INT_SIZE, 1, f );
	       FWRITE( VSliceTable[ip][it].vx3, INT_2_SIZE, num3, f );
	       FWRITE( VSliceTable[ip][it].vy3, INT_2_SIZE, num3, f );
	       FWRITE( VSliceTable[ip][it].vz3, INT_2_SIZE, num3, f );
	    }
	 }
      }
   } /* just colors */

   /* save vertical contour slice colors */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      tag = TAG_VSLICE_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      FWRITE( &ip, INT_SIZE, 1, f );
      FWRITE( &Color[ip][VSLICE], ULONG_SIZE, 1, f );
   }

   if (!justcolors) {
      /* save horizontal colored slices */
      for (ip=0;ip<NumParms+NumClones;ip++) {
	 for (it=0;it<NumTimes;it++) {
	    if (CHSliceTable[ip][it].valid) {
	       int num;
	       tag = TAG_CHSLICE;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       num = Nr * Nc;
	       bytes = 2*INT_SIZE + FLOAT_SIZE + INT_SIZE + num * UINT_1_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       FWRITE( &ip, INT_SIZE, 1, f );  
	       FWRITE( &it, INT_SIZE, 1, f );  
	       FWRITE( &CHSliceTable[ip][it].level, FLOAT_SIZE, 1, f );
	       FWRITE( &num, INT_SIZE, 1, f );
	       FWRITE( CHSliceTable[ip][it].data, UINT_1_SIZE, num, f );
	    }
	 }
      }
   }

   /* save horizontal color slice colors */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      tag = TAG_CHSLICE_COLORS;
      FWRITE( &tag, TAG_SIZE, 1, f );
      size = 256;
      bytes = 2*INT_SIZE + ULONG_SIZE + size*ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      /* CHSLICE colors */
      FWRITE( &ip, INT_SIZE, 1, f );  /* parm */
      FWRITE( &Color[ip][CHSLICE], ULONG_SIZE, 1, f );
      FWRITE( &size, INT_SIZE, 1, f );
      FWRITE( CHSliceColors[ip], ULONG_SIZE, size, f );
   }

   if (!justcolors) {
      /* save vertical color slices */
      for (ip=0;ip<NumParms+NumClones;ip++) {
	 for (it=0;it<NumTimes;it++) {
	    if (CVSliceTable[ip][it].valid) {
	       int num;
	       tag = TAG_CVSLICE;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       num = CVSliceTable[ip][it].rows * CVSliceTable[ip][it].cols;
	       bytes = 4*INT_SIZE + 4*FLOAT_SIZE + num*UINT_1_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       FWRITE( &ip, INT_SIZE, 1, f );  
	       FWRITE( &it, INT_SIZE, 1, f );  
	       FWRITE( &CVSliceTable[ip][it].r1, FLOAT_SIZE, 1, f );
	       FWRITE( &CVSliceTable[ip][it].c1, FLOAT_SIZE, 1, f );
	       FWRITE( &CVSliceTable[ip][it].r2, FLOAT_SIZE, 1, f );
	       FWRITE( &CVSliceTable[ip][it].c2, FLOAT_SIZE, 1, f );
	       FWRITE( &CVSliceTable[ip][it].rows, INT_SIZE, 1, f );
	       FWRITE( &CVSliceTable[ip][it].cols, INT_SIZE, 1, f );
	       FWRITE( CVSliceTable[ip][it].data, UINT_1_SIZE, num, f );
	    }
	 }
      }
   }

   /* save vertical color slice colors */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      tag = TAG_CVSLICE_COLORS;
      FWRITE( &tag, TAG_SIZE, 1, f );
      size = 256;
      bytes = 2*INT_SIZE + ULONG_SIZE + size*ULONG_SIZE;
      FWRITE( &bytes, 4, 1, f );
      /* CVSLICE colors */
      FWRITE( &ip, INT_SIZE, 1, f );  /* parm */
      FWRITE( &Color[ip][CVSLICE], ULONG_SIZE, 1, f );
      FWRITE( &size, INT_SIZE, 1, f );
      FWRITE( CVSliceColors[ip], ULONG_SIZE, size, f );
   }
   
   if (!justcolors) {
      /* save trajectories */
      for (i=0;i<NumTraj;i++) {
	 int length = TrajTable[i].length;
	 tag = TAG_TRAJ;
	 FWRITE( &tag, TAG_SIZE, 1, f );
	 bytes = 4 * INT_SIZE
	       + length * 3 * INT_2_SIZE
	       + length * 3 * INT_1_SIZE * (TrajTable[i].kind==1)
	       + NumTimes * 2 * UINT_2_SIZE;
	 FWRITE( &bytes, INT_SIZE, 1, f );
	 FWRITE( &TrajTable[i].length, INT_SIZE, 1, f );
	 FWRITE( &TrajTable[i].group, INT_SIZE, 1, f );
	 FWRITE( &TrajTable[i].kind, INT_SIZE, 1, f );
	 FWRITE( &NumTimes, INT_SIZE, 1, f );
	 FWRITE( TrajTable[i].vx, INT_2_SIZE, length, f );
	 FWRITE( TrajTable[i].vy, INT_2_SIZE, length, f );
	 FWRITE( TrajTable[i].vz, INT_2_SIZE, length, f );
	 if (TrajTable[i].kind==1) {
	    FWRITE( TrajTable[i].nx, INT_1_SIZE, length, f );
	    FWRITE( TrajTable[i].ny, INT_1_SIZE, length, f );
	    FWRITE( TrajTable[i].nz, INT_1_SIZE, length, f );
	 }
	 FWRITE( TrajTable[i].start, UINT_2_SIZE, NumTimes, f );
	 FWRITE( TrajTable[i].len, UINT_2_SIZE, NumTimes, f );
      }
      
   } /* just colors */
   
   /* save traj colors */
   for (i=0;i<TRAJGROUPS;i++) {
      tag = TAG_TRAJ_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      FWRITE( &i, INT_SIZE, 1, f );
      FWRITE( &TrajColor[i], ULONG_SIZE, 1, f );
   }

   /* save horizontal wind slices */
   if (!justcolors) {
      int ws;
      for (ws=0;ws<WINDSLICES;ws++) {
	 for (it=0;it<NumTimes;it++) {
	    if (HWindTable[ws][it].valid) {
	       int nvect;
	       tag = TAG_HWIND;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       nvect = HWindTable[ws][it].nvectors;
	       bytes = 3*INT_SIZE + 3*FLOAT_SIZE + 4*3*nvect*INT_2_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       FWRITE( &ws, INT_SIZE, 1, f );
	       FWRITE( &it, INT_SIZE, 1, f );
	       FWRITE( &HWindTable[ws][it].level, FLOAT_SIZE, 1, f );
	       FWRITE( &HWindTable[ws][it].density, FLOAT_SIZE, 1, f );
	       FWRITE( &HWindTable[ws][it].scale, FLOAT_SIZE, 1, f );
	       FWRITE( &nvect, INT_SIZE, 1, f );
	       FWRITE( HWindTable[ws][it].vx, INT_2_SIZE, 4*nvect, f );
	       FWRITE( HWindTable[ws][it].vy, INT_2_SIZE, 4*nvect, f );
	       FWRITE( HWindTable[ws][it].vz, INT_2_SIZE, 4*nvect, f );
	    }
	 }
      }
   }

   /* save horizontal wind slice colors */
   for (i=0;i<WINDSLICES;i++) {
      tag = TAG_HWIND_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      FWRITE( &i, INT_SIZE, 1, f );
      FWRITE( &HWindColor[i], ULONG_SIZE, 1, f );
   }

   /* save vertical wind slices */
   if (!justcolors) {
      int ws;
      for (ws=0;ws<WINDSLICES;ws++) {
	 for (it=0;it<NumTimes;it++) {
	    if (VWindTable[ws][it].valid) {
	       int nvect;
	       tag = TAG_VWIND;
	       FWRITE( &tag, TAG_SIZE, 1, f );
	       nvect = VWindTable[ws][it].nvectors;
	       bytes = 3*INT_SIZE + 6*FLOAT_SIZE + nvect*4*3*INT_2_SIZE;
	       FWRITE( &bytes, INT_SIZE, 1, f );
	       FWRITE( &ws, INT_SIZE, 1, f );
	       FWRITE( &it, INT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].r1, FLOAT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].c1, FLOAT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].r2, FLOAT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].c2, FLOAT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].density, FLOAT_SIZE, 1, f );
	       FWRITE( &VWindTable[ws][it].scale, FLOAT_SIZE, 1, f );
	       FWRITE( &nvect, INT_SIZE, 1, f );
	       FWRITE( VWindTable[ws][it].vx, INT_2_SIZE, 4*nvect, f );
	       FWRITE( VWindTable[ws][it].vy, INT_2_SIZE, 4*nvect, f );
	       FWRITE( VWindTable[ws][it].vz, INT_2_SIZE, 4*nvect, f );
	    }
	 }
      }
   }

   /* save vertical wind slice colors */
   for (i=0;i<WINDSLICES;i++) {
      tag = TAG_VWIND_COLOR;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = INT_SIZE + ULONG_SIZE;
      FWRITE( &bytes, INT_SIZE, 1, f );
      FWRITE( &i, INT_SIZE, 1, f );
      FWRITE( &VWindColor[i], ULONG_SIZE, 1, f );
   }

   if (!justcolors) {
      /* save CTM */
      tag = TAG_CTM;
      FWRITE( &tag, TAG_SIZE, 1, f );
      bytes = FLOAT_SIZE * 16;
      FWRITE( &bytes, 4, 1, f );
      FWRITE( CTM, FLOAT_SIZE, 16, f );
      
      /* save text labels */
      for (lab=FirstLabel;lab;lab=lab->next) {
	 tag = TAG_LABEL;
	 FWRITE( &tag, TAG_SIZE, 1, f );
	 bytes = strlen(lab->text)+1 + 3*INT_SIZE;
	 FWRITE( &bytes, 4, 1, f );
	 FWRITE( &lab->len, INT_SIZE, 1, f );     /* length of label */
	 FWRITE( lab->text, lab->len+1, 1, f );   /* label text + '\0' */
	 FWRITE( &lab->x, INT_SIZE, 1, f );       /* position */
	 FWRITE( &lab->y, INT_SIZE, 1, f );       /* position */
      }
   } /* just colors */

   /* save topography color table */
   tag = TAG_TOPO_COLORS;
   FWRITE( &tag, TAG_SIZE, 1, f );
   bytes = 256 * ULONG_SIZE;
   FWRITE( &bytes, INT_SIZE, 1, f );
   FWRITE( topo_color_table, ULONG_SIZE, 256, f );

   LOCK_OFF( Gfx_Lock );

   fclose(f);
   return 0;

}





/*** restore **********************************************************
   Restore the graphics saved by save() above.
   Input:  savefile - filename to restore from.
   Return:  0 for success,
            1 if unable to open file
	    2 if error while reading file.
**********************************************************************/
int restore( savefile )
char savefile[];
{
   FILE *f;
   int tag, blocklength;
   int xlate[1000];
/*   int bytes;*/
/*   float interval, r1,c1, r2,c2;*/

   f = fopen(savefile,"r");
   if (!f)
     return 1;


   while (!feof(f)) {
      if (fread( &tag, 1, INT_SIZE, f) <= 0)
	break;
      if (fread( &blocklength, 1, INT_SIZE, f ) <= 0)
	break;
      switch (tag) {
	 case TAG_CLONE_TABLE: {
	    int saved_clone_table[1000], flag[1000]; /* flag[i] is TRUE if */
	                       /* there exists a j such that xlate[j] = i  */
	    int i, j, size, found;

	    fread( &size, INT_SIZE, 1, f );
	    fread( saved_clone_table, INT_SIZE, size, f );

	    /* make a variable translation table such that when */
	    /* a graphic is restored, we can determine which */
	    /* variable the graphic corresponds to in the current */
	    /* clone table */
	    for (j=0;j<1000;j++)
	      flag[j] = 0;

	    for (i=0;i<size;i++) {
	       /* look if variable exists */
	       found = 0;
	       for (j=0;j<NumParms+NumClones;j++) {
		  if (CloneTable[j]==saved_clone_table[i] && !flag[j]) {
		     /* found */
		     found = 1;
		     xlate[i] = j;
		     flag[j] = 1;
		     break;
		  }
	       }
	       if (!found) {
		  /* make a new clone */
		  j = make_clone( saved_clone_table[i] );
		  xlate[i] = j;
		  if (j>=0) {
		     flag[j] = 1;
		  }
	       }
	    }
	 } break;

         case TAG_ISOSURF: {
	    int ip, it, numverts, numpts;
	    float level;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	      skip( f, blocklength-INT_SIZE );
	    fread( &it, INT_SIZE, 1, f );
	    fread( &level, FLOAT_SIZE, 1, f );
	    fread( &numverts, INT_SIZE, 1, f );
	    fread( &numpts, INT_SIZE, 1, f );
	    if (ip>=NumParms+NumClones || it>=NumTimes) {
	       skip(f, numverts*3*(INT_2_SIZE+INT_1_SIZE)+numpts*UINT_2_SIZE);
	       break; /*exit switch*/
	    }
	    recent( ISOSURF, ip );
	    LOCK_ON( Gfx_Lock );

	    /* deallocate old surface, if any */
	    free_isosurface( it, ip );

	    /* read isosurface */
	    SurfTable[ip][it].vx = (int_2 *) allocate( numverts*INT_2_SIZE );
	    if (fread(SurfTable[ip][it].vx,INT_2_SIZE,numverts,f)!=numverts) {
	       printf("bad read\n");
	    }
	    SurfTable[ip][it].vy = (int_2 *) allocate( numverts*INT_2_SIZE );
	    fread( SurfTable[ip][it].vy, INT_2_SIZE, numverts, f );
	    SurfTable[ip][it].vz = (int_2 *) allocate( numverts*INT_2_SIZE );
	    fread( SurfTable[ip][it].vz, INT_2_SIZE, numverts, f );
	    SurfTable[ip][it].nx = (int_1 *) allocate( numverts*INT_1_SIZE );
	    fread( SurfTable[ip][it].nx, INT_1_SIZE, numverts, f );
	    SurfTable[ip][it].ny = (int_1 *) allocate( numverts*INT_1_SIZE );
	    fread( SurfTable[ip][it].ny, INT_1_SIZE, numverts, f );
	    SurfTable[ip][it].nz = (int_1 *) allocate( numverts*INT_1_SIZE );
	    fread( SurfTable[ip][it].nz, INT_1_SIZE, numverts, f );
	    SurfTable[ip][it].pts = (uint_2 *) allocate( numpts*UINT_2_SIZE );
	    fread( SurfTable[ip][it].pts, UINT_2_SIZE, numpts, f );
	    SurfTable[ip][it].isolevel = level;
	    SurfTable[ip][it].numverts = numverts;
	    SurfTable[ip][it].numpts = numpts;
	    SurfTable[ip][it].valid = 1;
	    IsoLevel[ip] = level;
	     
	    LOCK_OFF( Gfx_Lock );
	 } break;
	 
         case TAG_ISO_COLOR: {
	    int ip;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    else
	       fread( &Color[ip][ISOSURF], ULONG_SIZE, 1, f );
	 } break;

	 case TAG_HSLICE: {
	    int ip, it, num1, num2, num3;
	    float interval, level;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &it, INT_SIZE, 1, f );
	    fread( &interval, FLOAT_SIZE, 1, f );
	    fread( &level, FLOAT_SIZE, 1, f );
	    
	    recent( HSLICE, ip );
	    LOCK_ON( Gfx_Lock );
	    /* deallocate old slice, if any */
	    free_hslice( it, ip );
	    
	    fread( &num1, INT_SIZE, 1, f );
	    HSliceTable[ip][it].num1 = num1;
	    HSliceTable[ip][it].vx1 = (int_2 *) allocate( num1 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vx1, INT_2_SIZE, num1, f );
	    HSliceTable[ip][it].vy1 = (int_2 *) allocate( num1 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vy1, INT_2_SIZE, num1, f );
	    
	    fread( &num2, INT_SIZE, 1, f );
	    HSliceTable[ip][it].num2 = num2;
	    HSliceTable[ip][it].vx2 = (int_2 *) allocate( num2 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vx2, INT_2_SIZE, num2, f );
	    HSliceTable[ip][it].vy2 = (int_2 *) allocate( num2 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vy2, INT_2_SIZE, num2, f );
	    
	    fread( &num3, INT_SIZE, 1, f );
	    HSliceTable[ip][it].num3 = num3;
	    HSliceTable[ip][it].vx3 = (int_2 *) allocate( num3 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vx3, INT_2_SIZE, num3, f );
	    HSliceTable[ip][it].vy3 = (int_2 *) allocate( num3 * INT_2_SIZE );
	    fread( HSliceTable[ip][it].vy3, INT_2_SIZE, num3, f );
	    
	    HSliceTable[ip][it].level = level;
	    HSliceTable[ip][it].interval = interval;
	    HSliceTable[ip][it].valid = 1;
	    HSliceLevel[ip] = level;
	    HSliceInterval[ip] = interval;

	    LOCK_OFF( Gfx_Lock );
	 } break;

         case TAG_HSLICE_COLOR: {
	    int ip;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    if (ip<NumParms+NumClones)
	       fread( &Color[ip][HSLICE], ULONG_SIZE, 1, f );
	    else
	       skip( f, ULONG_SIZE );
	 } break;

         case TAG_VSLICE: {
	    int ip, it, num1, num2, num3;
	    float interval, r1,c1, r2,c2;
	     
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &it, INT_SIZE, 1, f );
	    fread( &interval, FLOAT_SIZE, 1, f );
	    fread( &r1, FLOAT_SIZE, 1, f );
	    fread( &c1, FLOAT_SIZE, 1, f );
	    fread( &r2, FLOAT_SIZE, 1, f );
	    fread( &c2, FLOAT_SIZE, 1, f );
	     
	    recent( VSLICE, ip );
	    LOCK_ON( Gfx_Lock );

	    /* deallocate old slice, if any */
	    free_vslice( it, ip );
	     
	    fread( &num1, INT_SIZE, 1, f );
	    VSliceTable[ip][it].num1 = num1;
	    VSliceTable[ip][it].vx1 = (int_2 *) allocate( num1 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vx1, INT_2_SIZE, num1, f );
	    VSliceTable[ip][it].vy1 = (int_2 *) allocate( num1 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vy1, INT_2_SIZE, num1, f );
	    VSliceTable[ip][it].vz1 = (int_2 *) allocate( num1 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vz1, INT_2_SIZE, num1, f );
	     
	    fread( &num2, INT_SIZE, 1, f );
	    VSliceTable[ip][it].num2 = num2;
	    VSliceTable[ip][it].vx2 = (int_2 *) allocate( num2 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vx2, INT_2_SIZE, num2, f );
	    VSliceTable[ip][it].vy2 = (int_2 *) allocate( num2 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vy2, INT_2_SIZE, num2, f );
	    VSliceTable[ip][it].vz2 = (int_2 *) allocate( num2 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vz2, INT_2_SIZE, num2, f );
	     
	    fread( &num3, INT_SIZE, 1, f );
	    VSliceTable[ip][it].num3 = num3;
	    VSliceTable[ip][it].vx3 = (int_2 *) allocate( num3 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vx3, INT_2_SIZE, num3, f );
	    VSliceTable[ip][it].vy3 = (int_2 *) allocate( num3 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vy3, INT_2_SIZE, num3, f );
	    VSliceTable[ip][it].vz3 = (int_2 *) allocate( num3 * INT_2_SIZE );
	    fread( VSliceTable[ip][it].vz3, INT_2_SIZE, num3, f );

	    VSliceTable[ip][it].r1 = r1;
	    VSliceTable[ip][it].c1 = c1;
	    VSliceTable[ip][it].r2 = r2;
	    VSliceTable[ip][it].c2 = c2;
	    VSliceTable[ip][it].interval = interval;
	    VSliceTable[ip][it].valid = 1;
	    VSliceR1[ip] = r1;
	    VSliceC1[ip] = c1;
	    VSliceR2[ip] = r2;
	    VSliceC2[ip] = c2;
	    VSliceInterval[ip] = interval;

	    LOCK_OFF( Gfx_Lock );
	 } break;

	 case TAG_VSLICE_COLOR: {
	    int ip;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    if (ip<NumParms+NumClones)
	       fread( &Color[ip][VSLICE], ULONG_SIZE, 1, f );
	    else
	       skip( f, ULONG_SIZE );
	 } break;

	 case TAG_CHSLICE: {
	    /* restore horizontal color slice */
	    int ip, it, num, bytes;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &it, INT_SIZE, 1, f );
	    recent( CHSLICE, ip );
	    LOCK_ON( Gfx_Lock );
	    /* deallocate old slice */
	    free_chslice( it, ip );

	    fread( &CHSliceTable[ip][it].level, FLOAT_SIZE, 1, f );
	    fread( &num, INT_SIZE, 1, f );
	    CHSliceTable[ip][it].data = (uint_1 *) allocate(num*UINT_1_SIZE);
	    fread( CHSliceTable[ip][it].data, UINT_1_SIZE, num, f );
	    CHSliceTable[ip][it].valid = 1;
	    CHSliceLevel[ip] = CHSliceTable[ip][it].level;
	    LOCK_OFF( Gfx_Lock );
	 } break;

         case TAG_CHSLICE_COLORS: {
	    int ip, size;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &Color[ip][CHSLICE], ULONG_SIZE, 1, f );
	    fread( &size, INT_SIZE, 1, f );
	    fread( CHSliceColors[ip], ULONG_SIZE, size, f );
	 } break;

	 case TAG_CVSLICE: {
	    /* restore vertical color slice */
	    int ip, it, rows, cols, bytes;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &it, INT_SIZE, 1, f );
	    recent( CVSLICE, ip );
	    LOCK_ON( Gfx_Lock );
	    /* deallocate old slice */
	    free_cvslice( it, ip );

	    fread( &CVSliceTable[ip][it].r1, FLOAT_SIZE, 1, f );
	    fread( &CVSliceTable[ip][it].c1, FLOAT_SIZE, 1, f );
	    fread( &CVSliceTable[ip][it].r2, FLOAT_SIZE, 1, f );
	    fread( &CVSliceTable[ip][it].c2, FLOAT_SIZE, 1, f );
	    fread( &rows, INT_SIZE, 1, f );
	    fread( &cols, INT_SIZE, 1, f );
	    bytes = rows * cols * UINT_1_SIZE;
	    CVSliceTable[ip][it].data = (uint_1 *) allocate(bytes);
	    fread( CVSliceTable[ip][it].data, 1, bytes, f );
	    CVSliceTable[ip][it].rows = rows;
	    CVSliceTable[ip][it].cols = cols;
	    CVSliceTable[ip][it].valid = 1;
	    CVSliceR1[ip] = CVSliceTable[ip][it].r1;
	    CVSliceC1[ip] = CVSliceTable[ip][it].c1;
	    CVSliceR2[ip] = CVSliceTable[ip][it].r2;
	    CVSliceC2[ip] = CVSliceTable[ip][it].c2;
	    LOCK_OFF( Gfx_Lock );
	 } break;

         case TAG_CVSLICE_COLORS: {
	    int ip, size;
	    fread( &ip, INT_SIZE, 1, f );
	    ip = xlate[ip];
	    if (ip<0)
	       skip( f, blocklength-INT_SIZE );
	    fread( &Color[ip][CVSLICE], ULONG_SIZE, 1, f );
	    fread( &size, INT_SIZE, 1, f );
	    fread( CVSliceColors[ip], ULONG_SIZE, size, f );
	 } break;

         case TAG_TRAJ: {
	    int length, numtimes, kind, group, bytes;
	    if (NumTraj<MAXTRAJ) {
	       fread( &length, INT_SIZE, 1, f );
	       fread( &group, INT_SIZE, 1, f );
	       recent( TRAJ, group );
	       fread( &kind, INT_SIZE, 1, f );
	       fread( &numtimes, INT_SIZE, 1, f );
	       TrajTable[NumTraj].length = length;
	       TrajTable[NumTraj].group = group;
	       TrajTable[NumTraj].kind = kind;
	       TrajTable[NumTraj].vx = (int_2 *) allocate(length*INT_2_SIZE);
	       TrajTable[NumTraj].vy = (int_2 *) allocate(length*INT_2_SIZE);
	       TrajTable[NumTraj].vz = (int_2 *) allocate(length*INT_2_SIZE);
	       fread( TrajTable[NumTraj].vx, INT_2_SIZE, length, f );
	       fread( TrajTable[NumTraj].vy, INT_2_SIZE, length, f );
	       fread( TrajTable[NumTraj].vz, INT_2_SIZE, length, f );
	       if (kind==1) {
		  /* read ribbon normals */
		  TrajTable[NumTraj].nx= (int_1*)allocate(length*UINT_1_SIZE);
		  TrajTable[NumTraj].ny= (int_1*)allocate(length*UINT_1_SIZE);
		  TrajTable[NumTraj].nz= (int_1*)allocate(length*UINT_1_SIZE);
		  fread( TrajTable[NumTraj].nx, INT_1_SIZE, length, f );
		  fread( TrajTable[NumTraj].ny, INT_1_SIZE, length, f );
		  fread( TrajTable[NumTraj].nz, INT_1_SIZE, length, f );
	       }
	       bytes = NumTimes * UINT_2_SIZE;
	       TrajTable[NumTraj].start = (uint_2 *) allocate(bytes);
	       TrajTable[NumTraj].len = (uint_2 *) allocate(bytes);
	       fread( TrajTable[NumTraj].start, UINT_2_SIZE, numtimes, f );
	       fread( TrajTable[NumTraj].len, UINT_2_SIZE, numtimes, f );
	       NumTraj++;
	    }
	    else {
	       skip( f, blocklength );
	    }
	 } break;
	 
         case TAG_TRAJ_COLOR: {
	    /* restore traj colors */
	    int grp;
	    fread( &grp, INT_SIZE, 1, f );
	    if (grp<TRAJGROUPS) {
	       fread( &TrajColor[grp], ULONG_SIZE, 1, f );
	    }
	    else {
	       skip( f, ULONG_SIZE );
	    }
	 } break;
	 
	 case TAG_CTM:
	    /* restore ctm */
	    fread( CTM, FLOAT_SIZE, 16, f );
	    mat_inv( CTMinv, CTM );
	    break;
	 
	 case TAG_LABEL: {
            /* text label */
            int len, x, y;
            char label[1000];
            fread( &len, INT_SIZE, 1, f );
	    fread( label, 1, len+1, f );
	    fread( &x, INT_SIZE, 1, f );
	    fread( &y, INT_SIZE, 1, f );
	    new_label( label, x, y );
	 } break;

         case TAG_HWIND: {
	    /* Horizontal wind slice */
	    int ws, it, nvect, bytes;
	    fread( &ws, INT_SIZE, 1, f );  /* which slice */
	    fread( &it, INT_SIZE, 1, f );  /* which time */
	    if (ws<WINDSLICES && it<NumTimes) {
	       recent( HWIND, ws );
	       LOCK_ON( Gfx_Lock );
	       free_hwind( it, ws );

	       fread( &HWindTable[ws][it].level, FLOAT_SIZE, 1, f );
	       fread( &HWindTable[ws][it].density, FLOAT_SIZE, 1, f );
	       fread( &HWindTable[ws][it].scale, FLOAT_SIZE, 1, f );
	       HWindLevel[ws] = HWindTable[ws][it].level;
	       HWindDensity[ws] = HWindTable[ws][it].density;
	       HWindScale[ws] = HWindTable[ws][it].scale;

	       fread( &nvect, INT_SIZE, 1, f );
	       HWindTable[ws][it].nvectors = nvect;
	       bytes = 4 * nvect * INT_2_SIZE;
	       HWindTable[ws][it].vx = (int_2 *) allocate(bytes);
	       fread( HWindTable[ws][it].vx, INT_2_SIZE, 4*nvect, f );
	       HWindTable[ws][it].vy = (int_2 *) allocate(bytes);
	       fread( HWindTable[ws][it].vy, INT_2_SIZE, 4*nvect, f );
	       HWindTable[ws][it].vz = (int_2 *) allocate(bytes);
	       fread( HWindTable[ws][it].vz, INT_2_SIZE, 4*nvect, f );
	       HWindTable[ws][it].valid = 1;
	       LOCK_OFF( Gfx_Lock );
	    }
	    else {
	       skip( f, blocklength - 2*INT_SIZE );
	    }
	 } break;

         case TAG_HWIND_COLOR: {
	    int ws;
	    fread( &ws, INT_SIZE, 1, f );
	    if (ws<WINDSLICES) {
	       fread( &HWindColor[ws], ULONG_SIZE, 1, f );
	    }
	    else {
	       skip(4, ULONG_SIZE);
	    }
	 } break;

         case TAG_VWIND: {
	    /* Vertical wind slice */
	    int ws, it, nvect, bytes;
	    fread( &ws, INT_SIZE, 1, f );  /* which slice */
	    fread( &it, INT_SIZE, 1, f );  /* which time */
	    if (ws<WINDSLICES && it<NumTimes) {
	       recent( VWIND, ws );
	       LOCK_ON( Gfx_Lock );
	       free_vwind( it, ws );

	       fread( &VWindTable[ws][it].r1, FLOAT_SIZE, 1, f );
	       fread( &VWindTable[ws][it].c1, FLOAT_SIZE, 1, f );
	       fread( &VWindTable[ws][it].r2, FLOAT_SIZE, 1, f );
	       fread( &VWindTable[ws][it].c2, FLOAT_SIZE, 1, f );
	       fread( &VWindTable[ws][it].density, FLOAT_SIZE, 1, f );
	       fread( &VWindTable[ws][it].scale, FLOAT_SIZE, 1, f );
	       VWindR1[ws] = VWindTable[ws][it].r1;
	       VWindC1[ws] = VWindTable[ws][it].c1;
	       VWindR2[ws] = VWindTable[ws][it].r2;
	       VWindC2[ws] = VWindTable[ws][it].c2;
	       VWindDensity[ws] = VWindTable[ws][it].density;
	       VWindScale[ws] = VWindTable[ws][it].scale;

	       fread( &nvect, INT_SIZE, 1, f );
	       VWindTable[ws][it].nvectors = nvect;
	       bytes = 4 * nvect * INT_2_SIZE;
	       VWindTable[ws][it].vx = (int_2 *) allocate(bytes);
	       fread( VWindTable[ws][it].vx, INT_2_SIZE, 4*nvect, f );
	       VWindTable[ws][it].vy = (int_2 *) allocate(bytes);
	       fread( VWindTable[ws][it].vy, INT_2_SIZE, 4*nvect, f );
	       VWindTable[ws][it].vz = (int_2 *) allocate(bytes);
	       fread( VWindTable[ws][it].vz, INT_2_SIZE, 4*nvect, f );
	       VWindTable[ws][it].valid = 1;
	       LOCK_OFF( Gfx_Lock );
	    }
	    else {
	       skip( f, blocklength - 2*INT_SIZE );
	    }
	 } break;

         case TAG_VWIND_COLOR: {
	    int ws;
	    fread( &ws, INT_SIZE, 1, f );
	    if (ws<WINDSLICES) {
	       fread( &VWindColor[ws], ULONG_SIZE, 1, f );
	    }
	    else {
	       skip(4, ULONG_SIZE);
	    }
	 } break;

	 case TAG_TOPO_COLORS:
	    fread( topo_color_table, ULONG_SIZE, 256, f );
	    recolor_topo( topo_color_table, 256 );
	    break;
	
	 default:
	    /* skip bytes */
	    printf("skipping block (tag=%d,len=%d)\n", tag, blocklength );
	    skip( f, blocklength );
      }
   }
   fclose(f);
   return 0;

}
